/********************************************************************
 *
 * getcolor.c
 *  - X-Programm zum Abfragen der Farbe eines Punktes auf dem Bildschirm
 *
 * Quelle: BREDEX GmbH, Braunschweig
 *
 ********************************************************************/

/* Header-Dateien */
#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>

/* fuer exklusiven Server-Zugriff:
 */
#define SERVER_GRAB

/* Vorwaerts-Deklarationen */
void selectPoint();
Window findWindow();
unsigned long getPixel();
void printFarbinfos();


main ()
{
    Display* display;             /* Display */
    int x, y;                     /* Root-Koordinaten des Punktes */
    Window w;                     /* Fenster, in dem sich der Punkt befindet */
    unsigned long pixel;          /* Farbwert vom Punkt */

    /* Display oeffnen (Verbindung zum X-Server herstellen)
     */
    display = XOpenDisplay ((char*)NULL);
    if (display == NULL) {
        fprintf (stderr, "can't open \"%s\"\n", XDisplayName((char*)NULL));
        exit (1);
    }

#ifdef SERVER_GRAB
    /* X-Server fuer Fremdzugriff sperren
     */
    XGrabServer (display);
#endif

    /* Punkt auf dem Bildschirm selektieren lassen
     */
    selectPoint (display, &x, &y);

    /* Fenster, in dem sich der Punkt befindet, ermitteln
     */
    w = findWindow (display, x, y);

    /* Farbwert vom Punkt ermitteln
     */
    pixel = getPixel (display, w, x, y);

    /* Farbwert ausgeben
     */
    printf ("Punkt (%d,%d) hat Farbwert: %d.\n", x, y, pixel);

    /* Farbinformationen ermitteln und ausgeben
     */
    printFarbinfos (display, w, pixel);

#ifdef SERVER_GRAB
    /* Server-Blockade aufheben
     */
    XUngrabServer (display);
#endif
}


/* selectPoint
 *  - Punkt auf dem Bildschirm selektieren lassen
 */
void selectPoint (display, xRet, yRet)
Display* display;        /* Display */
int* xRet;               /* Return-Wert fuer X-Koordinate */
int* yRet;               /* Return-Wert fuer Y-Koordinate */
{
    int status;
    Cursor cursor;
    XEvent event;

    /* Ziel-Cursor erzeugen
     */
    cursor = XCreateFontCursor (display, XC_crosshair);

    /* exklusiven Mauspointer-Zugriff mit Ziel-Cursor zum
     * Selektieren fordern
     */
    status = XGrabPointer (display, DefaultRootWindow(display), False,
                           ButtonPressMask|ButtonReleaseMask, GrabModeAsync,
                           GrabModeAsync, DefaultRootWindow(display),
                           cursor, CurrentTime);
    if (status != GrabSuccess) {
        fprintf (stderr, "can't grab the mouse");
        exit (1);
    }

    /* Schleife, bis Punkt selektiert
     */
    do {
        XNextEvent (display, &event);
    }
    while (event.type != ButtonPress);

    /* exklusiven Maus-Zugriff aufheben
     */
    XUngrabPointer (display, CurrentTime);

    /* absolute Mauskoordinaten als Return-Werte eintragen
     */
    *xRet = event.xbutton.x_root;
    *yRet = event.xbutton.y_root;
}


/* findWindow
 *  - Fenster zu Root-Koordinate zurueckliefern
 */
Window findWindow (display, x, y)
Display* display;       /* Display */
int x, y;               /* Root-Koordinaten */
{
    XWindowAttributes attr;
    Window findW;
    Window childW, newChildW;

    /* bei Root gehts los */
    findW = DefaultRootWindow(display);

    /* evtl. Child-Window fuer den Punkt ermitteln
     */
    XTranslateCoordinates (display, findW, findW,
                           x, y, &x, &y, &childW);

    /* solange Child-Window, in dem sich der Punkt befindet, existiert
     */
    while (childW != None) {
        /* Koordinaten-Transformation nach Child-Window
         * (dabei evtl. neues Child-Window ermitteln)
         */
        XTranslateCoordinates (display, findW, childW,
                               x, y, &x, &y, &newChildW);

        /* InputOnly-Fenster zaehlen nicht
         */
        if (newChildW != None) {
            XGetWindowAttributes (display, childW, &attr);
            if (attr.class != InputOutput) {
                return childW;
            }
        }

        /* fuer den naechsten Durchgang
         */
        findW = childW;
        childW = newChildW;
    }

    /* das ist es */
    return findW;
}


/* getPixel
 *  - Farbwert zu Punkt auf dem Display zurueckliefern
 */
unsigned long getPixel (display, w, xRoot, yRoot)
Display* display;       /* Display */
Window w;               /* Fenster, in dem sich der Punkt befindet */
int xRoot, yRoot;       /* Root-Koordinaten des Punktes */
{
  int xInWin, yInWin;           /* Fenster--Koordinaten des Punktes */
  Window childWin;              /* Dummy (nicht verwendet) */
  XImage* image;                /* Image zum Punkt */

  /* Koordinaten relativ zum Fenster umrechnen
   */
  XTranslateCoordinates (display, DefaultRootWindow(display), w,
                         xRoot, yRoot, &xInWin, &yInWin, &childWin);

  /* Ein-Punkt-Image zum Punkt im Fenster erzeugen
   */
  image = XGetImage (display, w, xInWin, yInWin,
                     1, 1, AllPlanes, ZPixmap);

  /* Farbwert daraus zureuckliefern
   */
  return XGetPixel (image, 0, 0);
}


/* printFarbinfos
 *  - Farbe(n) zum ubergebenen Farbwert ermitteln und ausgeben
 */
void printFarbinfos (display, w, pixel)
Display* display;            /* Display */
Window w;                    /* Fenster, in dem sich der Punkt befindet */
unsigned long pixel;         /* Farbwert vom Punkt */
{
    int screenNr = DefaultScreen(display);
    XWindowAttributes attr;    /* Fenster-Atribute */
    XColor color;              /* Struktur fuer Farbe zum Farbwert */

    /* Farben der Fenster-Colormap ausgeben
     */
    XGetWindowAttributes (display, w, &attr);
    color.pixel = pixel;
    XQueryColor (display, attr.colormap, &color);
    printf (" - RGB-Werte mit FENSTER-Colormap sind: (%x,%x,%x).\n",
                color.red, color.green, color.blue);

    /* Fenster-Colormap == Default-Colormap ?
     */
    if (attr.colormap == DefaultColormap(display,screenNr)) {
        printf (" - Punkt ist in Fenster mit Default-Colormap.\n");
    }
    else {
        printf (" - Punkt ist in Fenster mit EIGENER Colormap.\n");

        /* Farben der Default-Colormap ausgeben
         */
        XQueryColor (display, DefaultColormap(display,screenNr), &color);
        printf ("   RGB-Werte mit DEFAULT-Colormap sind: (%x,%x,%x).\n",
                    color.red, color.green, color.blue);
    }

    /* Fenster-Colormap installiert ?
     */
    if (attr.map_installed) {
        printf (" - Colormap vom Fenster ist installiert.\n");
    }
    else {
        Colormap* cmapList;    /* Liste der install. Colormaps */
        int numCmaps;          /* Anzahl der install. Colormaps */
        int i;

        printf (" - Colormap vom Fenster ist NICHT installiert.\n");

        /* Liste installierter Colormaps abfragen
         */
        cmapList = XListInstalledColormaps (display, w, &numCmaps);

        /* nur Default-Colormap installiert ?
         */
        if (numCmaps == 1 && cmapList[0] == DefaultColormap(display,screenNr)) {
            printf ("   Default-Colormap ist installiert.\n");
        }
        else {
            /* Farben der installierten Colormap(s) ausgeben
             */
            printf ("   RGB-Werte mit INSTALL. Colormap(s):");
            for (i=0; i<numCmaps; i++) {
                XQueryColor(display, cmapList[i], &color);
                printf (" (%x,%x,%x)", color.red, color.green, color.blue);
            }
            printf (".\n");
        }
    }
}

